<?php
/**
 * Generate class. Handles generating Pages, Posts, Custom Post Types
 * and Taxonomy Terms.
 * 
 * @package Page_Generator_Pro
 * @author  Tim Carr
 * @version 1.0.0
 */
class Page_Generator_Pro_Generate {

    /**
     * Holds the base object.
     *
     * @since   1.9.8
     *
     * @var     object
     */
    public $base;

    /**
     * Holds an array comprising of every keyword detected in the Group.
     * Each Keyword holds an array comprising of every single Term for that Keyword.
     *
     * @since   1.9.8
     *
     * @var     array
     */
    public $keywords = array();

    /**
     * Holds an array comprising of every keyword detected in the Group.
     * Each Keyword holds the nth Term that will be used to replace the Keyword.
     *
     * @since   1.9.8
     *
     * @var     array
     */
    public $keywords_terms = array();

    /**
     * Holds the array of found keywords across all settings.
     *
     * @since   1.2.0
     *
     * @var     array
     */
    public $required_keywords = array();

    /**
     * Holds the array of keywords to replace e.g. {city}
     *
     * @since   1.3.1
     *
     * @var     array
     */
    public $searches = array();

    /**
     * Holds the array of keyword values to replace e.g. Birmingham
     *
     * @since   1.3.1
     *
     * @var     array
     */
    public $replacements = array();

    /**
     * Holds a flag to denote whether Page Generator Pro shortcodes
     * should be processed on the main Post Content
     *
     * @since   1.9.5
     *
     * @var     bool
     */
    public $process_shortcodes_on_post_content = false;

    /**
     * Constructor.
     *
     * @since   1.9.3
     *
     * @param   object $base    Base Plugin Class
     */
    public function __construct( $base ) {

        // Store base class
        $this->base = $base;

        // Register Action Hooks
        add_action( 'page_generator_pro_generate_content_after', array( $this, 'generate_content_finished' ), 10, 2 );

    }

    /**
     * Calculates the maximum number of items that will be generated based
     * on the settings.
     *
     * @since   1.1.5
     *
     * @param   array   $settings   Group Settings (either a Content or Term Group)
     * @return  mixed               WP_Error | integer
     */
    public function get_max_number_of_pages( $settings ) {

        // Build a class array of required keywords that need replacing with data
        $this->find_keywords_in_settings( $settings );

        // Bail if no keywords were found
        if ( count( $this->required_keywords ) == 0 ) {
            return 0;
        }

        // Get the terms for each required keyword
        $this->keywords = $this->get_keywords_terms_columns_delimiters( $this->required_keywords );

        // Bail if no keywords were found
        if ( empty( $this->keywords['terms'] ) ) {
            return 0;
        }

        $total = 0;
        foreach ( $this->keywords['terms'] as $keyword => $terms ) {
            if ( count( $terms ) > 0 && ( count( $terms ) < $total || $total == 0 ) ) {
                $total = count( $terms );
            }
        }

        return $total;

    }
    
    /**
     * Main function to generate a Page, Post, Custom Post Type or Taxonomy Term
     *
     * @since   1.6.1
     *
     * @param   int     $group_id   Group ID
     * @param   int     $index      Keyword Index
     * @param   bool    $test_mode  Test Mode
     * @param   string  $system     System (browser|cron|cli)
     * @return  mixed               WP_Error | URL
     */
    public function generate_content( $group_id, $index = 0, $test_mode = false, $system = 'browser' ) {

        // Performance debugging
        $start  = microtime( true );

        // Define the Group ID and Index as globals, so it can be picked up by our shortcodes when they're processed
        global $page_generator_pro_group_id, $page_generator_pro_index;
        $page_generator_pro_group_id = $group_id;
        $page_generator_pro_index = $index;

        // If test mode is enabled, set the debug constant
        if ( $test_mode && ! defined( 'PAGE_GENERATOR_PRO_DEBUG' ) ) {
            define( 'PAGE_GENERATOR_PRO_DEBUG', true );
        }

        // If this Group has a request to cancel generation, exit
        if ( ! $test_mode ) {
            if ( $this->base->get_class( 'groups' )->cancel_generation_requested( $group_id ) ) {
                $this->base->get_class( 'groups' )->stop_generation( $group_id );
                return new WP_Error( 'generation_error', __( 'A request to cancel generation was made by the User. Exiting...', 'page-generator' ) );
            }
        }
        
        // Get group settings
        $settings = $this->base->get_class( 'groups' )->get_settings( $group_id );

        // If the Group is not published, generation might fail in Gutenberg stating that no keywords could be found
        // in the Content. Change its status to published
        if ( ! in_array( get_post_status( $group_id ), array_keys( $this->base->get_class( 'groups' )->get_group_statuses() ) ) ) {
            $result = wp_update_post( array(
                'ID'            => $group_id,
                'post_status'   => 'publish',
            ), true );

            if ( is_wp_error( $result ) ) {
                return $result;
            }
        }

        // Validate group
        $validated = $this->base->get_class( 'groups' )->validate( $group_id );
        if ( is_wp_error( $validated ) ) {
            return $validated;
        }

        /**
         * Run any actions before an individual Page, Post or Custom Post Type is generated
         * successfully.
         *
         * @since   2.4.1
         *
         * @param   int     $group_id       Group ID
         * @param   array   $settings       Group Settings
         * @param   int     $index          Keyword Index
         * @param   bool    $test_mode      Test Mode
         */
        do_action( 'page_generator_pro_generate_content_started', $group_id, $settings, $index, $test_mode );

        // Build a class array of required keywords that need replacing with data
        $this->find_keywords_in_settings( $settings );
        if ( count( $this->required_keywords ) == 0 ) {
            return new WP_Error( 'keyword_error', __( 'No keywords were specified in the Group.', 'page-generator' ) );
        }

        // Build a keywords array comprising of terms, columns and delimiters for each of the required keywords
        $this->keywords = $this->get_keywords_terms_columns_delimiters( $this->required_keywords );
        if ( count( $this->keywords['terms'] ) == 0 ) {
            return new WP_Error( 'keyword_error', __( 'Keywords were specified in the Group, but no keywords exist in either the Keywords section of the Plugin or as a Taxonomy.', 'page-generator' ) );
        }

        // Build array of keyword --> term key/value pairs to use for this generation
        $this->keywords_terms = $this->get_keywords_terms( $settings['method'], $index );
        if ( is_wp_error( $this->keywords_terms ) ) {
            return $this->keywords_terms;
        }

        // Rotate Author
        if ( isset( $settings['rotateAuthors'] ) ) {
            $authors = $this->base->get_class( 'common' )->get_authors();
            $user_index = ( $index % count( $authors ) );
        }

        // Iterate through each detected Keyword to build a full $this->searches and $this->replacements arrays
        $this->build_search_replace_arrays();
        
        // Iterate through each keyword and term key/value pair
        $settings = $this->replace_keywords( $settings );

        // Define Post Name / Slug
        // If no Permalink exists, use the Post Title
        if ( ! empty( $settings['permalink'] ) ) {
            $post_name = sanitize_title( $settings['permalink'] );
        } else {
            $post_name = sanitize_title( $settings['title'] );
        }

        /**
         * Modify the Group's settings prior to parsing shortcodes and building the Post Arguments
         * to use for generating a single Page, Post or Custom Post Type.
         *
         * Changes made only affect this item in the generation set, and are not persistent or saved.
         *
         * For Gutenberg and Page Builders with Blocks / Elements registered by this Plugin, this
         * is a good time to convert them to a Shortcode Block / Element / Text
         *
         * @since   2.6.0
         *
         * @param   array   $settings       Group Settings
         * @param   int     $group_id       Group ID
         * @param   int     $index          Keyword Index
         * @param   bool    $test_mode      Test Mode
         */
        $settings = apply_filters( 'page_generator_pro_generate_content_settings', $settings, $group_id, $index, $test_mode );

        // Build Post args
        $post_args = array(
            'post_type'     => $settings['type'],
            'post_title'    => $settings['title'],
            'post_content'  => $settings['content'],
            'post_status'   => ( $test_mode ? 'draft' : $settings['status'] ),
            'post_author'   => ( ( isset( $settings['rotateAuthors'] ) && $settings['rotateAuthors'] == 1 ) ? $authors[ $user_index ]->ID : $settings['author'] ), // ID
            'post_name'     => $post_name,
        );

        /**
         * Filters arguments used for creating or updating a Post when running
         * content generation.
         *
         * @since   1.6.1
         *
         * @param   array   $post_args  wp_insert_post() / wp_update_post() compatible arguments
         * @param   array   $settings   Content Group Settings
         */
        $post_args = apply_filters( 'page_generator_pro_generate_post_args', $post_args, $settings );

        /**
         * Run any actions immediately before an individual Page, Post or Custom Post Type is generated.
         *
         * @since   2.4.1
         *
         * @param   int     $group_id       Group ID
         * @param   array   $settings       Group Settings
         * @param   int     $index          Keyword Index
         * @param   bool    $test_mode      Test Mode
         */
        do_action( 'page_generator_pro_generate_content_before_insert_update_post', $group_id, $settings, $index, $test_mode );

        // Create Page
        $post_id = wp_insert_post( $post_args, true );

        // Define return message
        $log = sprintf( 
            __( 'Created', 'page-generator' ), 
            $settings['type']
        );

        // Check Post creation / update worked
        if ( is_wp_error( $post_id ) ) {
            // Fetch error codes when trying to insert / update the Post
            $error_codes = $post_id->get_error_codes();

            // Ignore invalid_page_template errors.  wp_update_post() adds the existing page_template
            // parameter to $post_args before passing onto wp_insert_post(); however the template
            // might belong to a Page Builder Template that has / will not register the template with
            // the active Theme.
            // We manually assign _wp_page_template later on in this process, so we can safely ignore
            // this error.
            if ( count( $error_codes ) == 1 && $error_codes[0] == 'invalid_page_template' ) {
                // The Post ID will be the existing Post ID we just updated
                $post_id = $existing_post_id;
            } else {
                // UTF-8 encode the Title and Content
                $post_args['post_title'] = utf8_encode( $post_args['post_title'] );
                $post_args['post_content'] = utf8_encode( $post_args['post_content'] );
                
                // Try again
                if ( isset( $post_args['ID'] ) ) {
                    // Update Page, Post or CPT
                    $post_id = wp_update_post( $post_args, true ); 
                } else {
                    // Create Page, Post or CPT
                    $post_id = wp_insert_post( $post_args, true ); 
                }

                // If Post creation / update still didn't work, bail
                if ( is_wp_error( $post_id ) ) {
                    $post_id->add_data( $post_args, $post_id->get_error_code() );
                    return $post_id;
                }
            }
        }

        /**
         * Run any actions immediately after an individual Page, Post or Custom Post Type is generated, but before
         * its Page Template, Featured Image, Custom Fields, Post Meta, Geodata or Taxonomy Terms have been assigned.
         *
         * @since   2.4.1
         *
         * @param   int     $post_id        Post ID
         * @param   int     $group_id       Group ID
         * @param   array   $settings       Group Settings
         * @param   int     $index          Keyword Index
         * @param   bool    $test_mode      Test Mode
         */
        do_action( 'page_generator_pro_generate_content_after_insert_update_post', $post_id, $group_id, $settings, $index, $test_mode );

        // Store this Group ID and Index in the Post's meta, so we can edit/delete the generated Post(s) in the future
        update_post_meta( $post_id, '_page_generator_pro_group', $group_id );
        update_post_meta( $post_id, '_page_generator_pro_index', $index );

        // Post / Page Template
        // Backward compat for Free
        if ( ! empty( $settings['pageTemplate'] ) && ! is_array( $settings['pageTemplate'] ) ) {
            update_post_meta( $post_id, '_wp_page_template', $settings['pageTemplate'] );
        }
        if ( ! empty( $settings['pageTemplate'][ $settings['type'] ] ) ) {
            update_post_meta( $post_id, '_wp_page_template', $settings['pageTemplate'][ $settings['type'] ] );
        }

        // Store Post Meta (ACF, Yoast, Page Builder data etc) on the Generated Post
        $this->set_post_meta( $post_id, $settings, $post_args );

        // Request that the user review the plugin. Notification displayed later,
        // can be called multiple times and won't re-display the notification if dismissed.
        if ( ! $test_mode ) {
            $this->base->dashboard->request_review();
        
            // Store current index as the last index generated for this Group, if we're not in test mode
            $this->base->get_class( 'groups' )->update_last_index_generated( $group_id, $index );
        }

        /**
         * Run any actions after an individual Page, Post or Custom Post Type is generated
         * successfully.
         *
         * @since   2.4.1
         *
         * @param   int     $post_id        Generated Post ID
         * @param   int     $group_id       Group ID
         * @param   array   $settings       Group Settings
         * @param   int     $index          Keyword Index
         * @param   bool    $test_mode      Test Mode
         */
        do_action( 'page_generator_pro_generate_content_finished', $post_id, $group_id, $settings, $index, $test_mode );

        // Return success data
        return $this->generate_return( $group_id, $post_id, $settings['type'], true, $log, $start, $test_mode, 'browser' );

    }

    /**
     * For all Keyword tags found in the Group, builds search and replacement class arrays for later use
     * when recursively iterating through a Group's settings to replace the Keyword tags with their Term counterparts
     *
     * @since   2.6.1
     */
    private function build_search_replace_arrays() {

        foreach ( $this->required_keywords_full as $keyword => $keywords_with_modifiers ) {
            // Build search and replacement arrays for this Keyword
            foreach ( $keywords_with_modifiers as $keyword_with_modifiers ) {
                // If the Keyword isn't truly a Keyword in the database, don't do anything
                if ( ! isset( $this->keywords_terms[ $keyword ] ) ) {
                    continue;
                }
                
                // Cast keyword as a string so numeric keywords don't break search/replace
                $this->build_search_replace_arrays_for_keyword( $keyword_with_modifiers, (string) $keyword, $this->keywords_terms[ $keyword ] );
            }
        }

    }

    /**
     * Appends the search and replace arrays for the given Keyword (column name, nth term, transformations) and its applicable Term.
     *
     * @since   2.6.1
     *
     * @param   string  $keyword_with_modifiers     Keyword with Modifiers (search, e.g. keyword(column):3:uppercase_all:url
     * @param   string  $keyword                    Keyword without Modifiers (e.g. keyword)
     * @param   string  $term                       Term (replacement)
     */
    private function build_search_replace_arrays_for_keyword( $keyword_with_modifiers, $keyword, $term ) {

        // If the Keyword with Modifiers matches the Keyword, we have no modifiers
        // Just return the term
        if ( $keyword_with_modifiers == $keyword ) {
            $this->searches[] = '{' . $keyword_with_modifiers . '}';
            $this->replacements[] = $term;
            return;
        }

    }

    /**
     * Helper method to iterate through each keyword's tags, including any modifiers,
     * building search and replacement arrays before recursively iterating through the supplied settings,
     * replacing the keywords and their transformations with the terms.
     *
     * @since   1.9.8
     *
     * @param   array   $settings   Group Settings
     * @return  array               Group Settings
     */
    public function replace_keywords( $settings ) {

        // Iterate through Group Settings, replacing $this->searches (Keywords) with $this->replacements (Terms)
        // as well as performing spintax and shortcode processing
        array_walk_recursive( $settings, array( $this, 'replace_keywords_in_array' ) );

        // Return
        return $settings;

    }

    /**
     * Recursively goes through the settings array, finding any {keywords}
     * specified, to build up an array of keywords we need to fetch.
     *
     * @since   1.0.0
     *
     * @param   array   $settings   Settings
     */
    private function find_keywords_in_settings( $settings ) {

        // Recursively walk through all settings to find all keywords
        array_walk_recursive( $settings, array( $this, 'find_keywords_in_string' ) );

    }

    /**
     * For the given array of keywords, only returns keywords with terms.
     *
     * @since   1.6.5
     *
     * @param   array   $required_keywords  Required Keywords
     * @return  array                       Keywords with Terms
     */
    private function get_keywords_terms_columns_delimiters( $required_keywords ) {

        // Define blank array for keywords with terms and keywords with columns
        $results = array(
            'terms'     => array(),
        );

        foreach ( $required_keywords as $key => $keyword ) {
            $result = $this->base->get_class( 'keywords' )->get_by( 'keyword', $keyword );

            // Skip if no results
            if ( ! is_array( $result ) ) {
                continue;
            }
            if ( count( $result ) == 0 ) {
                continue;
            }

            $results['terms'][ $keyword ] = $result['dataArr'];
        }

        // Return results
        return $results;

    }

    /**
     * Returns an array of keyword and term key / value pairs.
     * 
     * @since   1.0.0
     *
     * @param   string  $method     Generation Method
     * @param   int     $index      Generation Index
     * @return  mixed               WP_Error | array
     * 
     */
    private function get_keywords_terms( $method, $index ) {

        $keywords_terms = array();
        foreach ( $this->keywords['terms'] as $keyword => $terms ) {
            // Use modulo to get the term index for this keyword
            $term_index = ( $index % count( $terms ) );   

            // Build the keyword => term key/value pairs
            $keywords_terms[ $keyword ] = $terms[ $term_index ];
        }

        // Cleanup the terms
        foreach ( $keywords_terms as $key => $term ) {
            $keywords_terms[ $key ] = trim( html_entity_decode( $term ) );
        }

        return $keywords_terms;

    }

    /**
     * Performs a search on the given string to find any {keywords}
     *
     * @since 1.2.0
     *
     * @param   string  $content    Array Value (string to search)
     * @param   string  $key        Array Key
     */
    private function find_keywords_in_string( $content, $key ) {

        // If $content is an object, iterate this call
        if ( is_object( $content ) ) {
            return array_walk_recursive( $content, array( $this, 'find_keywords_in_string' ) );
        }

        // Get keywords and spins in this string
        preg_match_all( "|{(.+?)}|", $content, $matches );

        // Bail if no matches are found
        if ( ! is_array( $matches ) ) {
            return;
        }
        if ( count( $matches[1] ) == 0 ) {
            return;
        }

        // Iterate through matches
        foreach ( $matches[1] as $m_key => $keyword ) {
            // If this is a spin, ignore it
            if ( strpos( $keyword, "|" ) !== false ) {
                continue;
            }

            // If a keyword is within spintax at the start of the string (e.g. {{service}|{service2}} ),
            // we get an additional leading curly brace for some reason.  Remove it
            $keyword = str_replace( '{', '', $keyword );
            $keyword = str_replace( '}', '', $keyword );

            // Lowercase keyword, to avoid duplicates e.g. {City} and {city}
            $keyword = strtolower( $keyword );

            // Fetch just the Keyword Name
            $keyword_name = $this->extract_keyword_name_from_keyword( $keyword );

            // If the Keyword Name is not in our required_keywords array, add it now
            if ( ! in_array( $keyword_name, $this->required_keywords ) ) {
                $this->required_keywords[ $keyword_name ] = $keyword_name;
            }

            // If the Keyword (Full) is not in our required_keywords_full array, add it now
            if ( ! isset( $this->required_keywords_full[ $keyword_name ] ) ) {
                $this->required_keywords_full[ $keyword_name ] = array();
            }
            if ( ! in_array( $keyword, $this->required_keywords_full[ $keyword_name ] ) ) {
                $this->required_keywords_full[ $keyword_name ][] = $keyword;
            }
        }

    }

    /**
     * Returns just the keyword name, excluding any columns, nth terms and transformations
     *
     * @since   2.6.1
     *
     * @param   string  $keyword    Keyword
     * @return  string              Keyword Name excluding any columns, nth terms and transformations
     */
    private function extract_keyword_name_from_keyword( $keyword ) {

        if ( strpos( $keyword, ':' ) !== false ) {
            $keyword_parts = explode( ':', $keyword );
            $keyword = trim( $keyword_parts[0] );
        }

        $keyword = preg_replace( '/\(.*?\)/', '', $keyword );

        return $keyword;

    }

    /**
     * array_walk_recursive callback, which finds $this->searches, replacing with
     * $this->replacements in $item.
     *
     * Also performs spintax.
     *
     * @since   1.3.1
     *
     * @param   mixed   $item   Item (array, object, string)
     * @param   string  $key    Key
     */
    private function replace_keywords_in_array( &$item, $key ) {

        // If the settings key's value is an array, walk through it recursively to search/replace
        // Otherwise do a standard search/replace on the string
        if ( is_array( $item ) ) {
            // Array
            array_walk_recursive( $item, array( $this, 'replace_keywords_in_array' ) );
        } elseif( is_object( $item ) ) {
            // Object
            array_walk_recursive( $item, array( $this, 'replace_keywords_in_array' ) );
        } elseif ( is_string( $item ) ) {
            // If here, we have a string
            // Perform keyword replacement, spintax and shortcode processing now
            $item = str_ireplace( $this->searches, $this->replacements, $item );

            /**
             * Perform any other keyword replacements or string processing.
             *
             * @since   1.9.8
             *
             * @param   string  $item   Group Setting String (this can be Post Meta, Custom Fields, Permalink, Title, Content etc)
             * @param   string  $key    Group Setting Key
             */
            $item = apply_filters( 'page_generator_pro_generate_replace_keywords_in_array', $item, $key );
        }

    }

/**
     * Copies the Content Group's Post Meta to the Generated Post ID,
     * including Page Builder / ACF data.
     *
     * @since   2.3.5
     *
     * @param   int     $post_id    Generated Post ID
     * @param   array   $settings   Group Settings
     * @param   array   $post_args  wp_insert_post() / wp_update_post() arguments
     * @return  bool                Updated Post Meta on Generated Post ID
     */
    private function set_post_meta( $post_id, $settings, $post_args ) {

        // Bail if no Post Meta to copy to the generated Post
        if ( ! isset( $settings['post_meta'] ) ) {
            return false;
        }

        // Define the metadata to ignore
        $ignored_keys = array(
            '_wp_page_template',
        );

        /**
         * Defines Post Meta Keys in a Content Group to ignore and not copy to generated Posts / Groups.
         *
         * @since   2.6.1
         *
         * @param   array   $ignored_keys   Ignored Keys
         */
        $ignored_keys = apply_filters( 'page_generator_pro_generate_set_post_meta_ignored_keys', $ignored_keys );

        // Iterate through Post Meta
        foreach ( $settings['post_meta'] as $meta_key => $meta_value ) {

            // Skip ignored keys
            if ( in_array( $meta_key, $ignored_keys ) ) {
                continue;
            }

            /**
             * Filters the Group Metadata for the given Key and Value, immediately before it's
             * saved to the Generated Page.
             *
             * @since   2.6.1
             *
             * @param   mixed   $value  Meta Value
             * @return  mixed           Meta Value
             */
            $meta_value = apply_filters( 'page_generator_pro_generate_set_post_meta_' . $meta_key, $meta_value );

            // Update Generated Page's Meta Value
            update_post_meta( $post_id, $meta_key, $meta_value );

        }

        return true;
        
    }

    /**
     * Runs any actions once content generation has completed.
     *
     * @since   1.9.3
     *
     * @param   int     $group_id   Group ID
     * @param   bool    $test_mode  Test Mode
     */
    public function generate_content_finished( $group_id, $test_mode ) {

        include_once( ABSPATH . 'wp-admin/includes/plugin.php' );

        // Elementor: Clear Cache
        if ( is_plugin_active( 'elementor/elementor.php' ) ) {
            Elementor\Plugin::$instance->files_manager->clear_cache();
        }

    }

    /**
     * Main function to trash previously generated Contents
     * for the given Group ID
     *
     * @since   1.2.3
     *
     * @param   int     $group_id   Group ID
     * @return  mixed               WP_Error | Success
     */
    public function trash_content( $group_id ) {

        // Get all Post IDs generated by this Group
        $post_ids = $this->get_generated_content_post_ids( $group_id );

        // Bail if an error occured
        if ( is_wp_error( $post_ids ) ) {
            return $post_ids;
        }

        // Delete Posts by their IDs
        foreach ( $post_ids as $post_id ) {
            $result = wp_trash_post( $post_id );
            if ( ! $result ) {
                return new WP_Error( 'page_generator_pro_generate_trash_content', __( 'Unable to trash generated content with ID = ' . $post_id, 'page-generator' ) );
            }
        }

        // Done
        return true;

    }

    /**
     * Main function to delete previously generated Contents
     * for the given Group ID
     *
     * @since   1.2.3
     *
     * @param   int     $group_id   Group ID
     * @return  mixed               WP_Error | Success
     */
    public function delete_content( $group_id ) {

        // Get all Post IDs generated by this Group
        $post_ids = $this->get_generated_content_post_ids( $group_id );

        // Bail if an error occured
        if ( is_wp_error( $post_ids ) ) {
            return $post_ids;
        }

        // Delete Attachments
        $this->delete_attachments_by_post_ids( $post_ids, $group_id );

        // Delete Posts by their IDs
        foreach ( $post_ids as $post_id ) {
            $result = wp_delete_post( $post_id, true );
            if ( ! $result ) {
                return new WP_Error( 'page_generator_pro_generate_delete_content', __( 'Unable to delete generated content with ID = ' . $post_id, 'page-generator' ) );
            }
        }

        // Done
        return true;

    }

    /**
     * Returns all Post IDs generated by the given Group ID
     *
     * @since   1.9.1
     *
     * @param   int     $group_id   Group ID
     * @return  mixed               WP_Error | array
     */
    private function get_generated_content_post_ids( $group_id ) {

        // Fetch valid Post Statuses that can be used when generating content
        $statuses = array_keys( $this->base->get_class( 'common' )->get_post_statuses() );

        // Get all Posts
        $posts = new WP_Query( array (
            'post_type'     => 'any',
            'post_status'   => $statuses,
            'posts_per_page'=> -1,
            'meta_query'    => array(
                array(
                    'key'   => '_page_generator_pro_group',
                    'value' => absint( $group_id ),
                ),
            ),
            'fields'        => 'ids',
        ) );

        // If no Posts found, return an error
        if ( count( $posts->posts ) == 0 ) {
            return new WP_Error( 'page_generator_pro_generate_get_generated_content_post_ids', __( 'No content has been generated by this group.', 'page-generator' ) );
        }

        // Return Post IDs
        return $posts->posts;

    }

    /**
     * Deletes Attachments assigned to the given Post ID and Group ID
     *
     * @since   2.4.1
     *
     * @param   array   $post_ids   Post IDs
     * @param   int     $group_id   Group ID
     * @param   int     $index      Generation Index
     * @return  mixed               WP_Error | bool
     */
    private function delete_attachments_by_post_ids( $post_ids, $group_id ) {

        // Build query
        $args = array(
            'post_type'         => 'attachment',
            'post_status'       => 'any',
            'posts_per_page'    => -1,
            'post_parent__in'   => $post_ids,
            'meta_query'        => array(
                array(
                    'key'   => '_page_generator_pro_group',
                    'value' => absint( $group_id ),
                ),
            ),
            'fields'            => 'ids',
        );

        // Get all Attachments belonging to the given Post IDs
        $attachments = new WP_Query( $args );

        // If no Attachments found, return false, as there's nothing to delete
        if ( count( $attachments->posts ) == 0 ) {
            return false;
        }

        // Delete attachments
        foreach ( $attachments->posts as $attachment_id ) {
            wp_delete_attachment( $attachment_id );
        }

        return true;
        
    }

    /**
     * Returns an array of data relating to the successfully generated Post or Term,
     * logging the result if logging is enabled
     *
     * @since   2.1.8
     *
     * @param   int     $group_id               Group ID
     * @param   int     $post_or_term_id        Post or Term ID
     * @param   string  $post_type_or_taxonomy  Post Type or Taxonomy
     * @param   bool    $generated              Post Generated (false = skipped)
     * @param   string  $message                Message to return (created, updated, skipped etc.)
     * @param   int     $start                  Start Time
     * @param   bool    $test_mode              Test Mode
     * @param   string  $system                 System (browser|cron|cli)
     * @return                                  Success Data
     */
    private function generate_return( $group_id, $post_or_term_id, $post_type_or_taxonomy, $generated, $message, $start, $test_mode = false, $system = 'browser' ) {

        // Determine if we're returning data for a generated Post or Term
        // We check if it's a Taxonomy first as post_type_exists() fails for e.g. WooCommerce Products
        if ( taxonomy_exists( $post_type_or_taxonomy ) ) {
            // Term
            $url = get_term_link( $post_or_term_id, $post_type_or_taxonomy );
        } else {
            // Post, Page or CPT
            if ( $test_mode ) {
                $url = get_bloginfo( 'url' ) . '?page_id=' . $post_or_term_id . '&preview=true';
            } else {
                $url = get_permalink( $post_or_term_id );
            }
        }

        // Performance debugging
        $end = microtime( true );

        // Build result array
        $result = array(
            // Item
            'post_id'           => $post_or_term_id,
            'url'               => $url,
            'type'              => ( taxonomy_exists( $post_type_or_taxonomy ) ? 'term' : 'content' ),
            'system'            => $system,
            'test_mode'         => $test_mode,
            'generated'         => $generated,
            'keywords_terms'    => $this->keywords_terms,
            'message'           => $message,

            // Performance data
            'start'             => $start,
            'end'               => $end,
            'duration'          => round ( ( $end - $start ), 2 ),
            'memory_usage'      => round( memory_get_usage() / 1024 / 1024 ),
            'memory_peak_usage' => round( memory_get_peak_usage() / 1024 / 1024 ),
        );

        // Maybe add to log
        if ( $this->base->get_class( 'settings' )->get_setting( $this->base->plugin->name . '-generate', 'log_enabled', '0' ) ) {
            $this->base->get_class( 'log' )->add( $group_id, $result );
        }

        // Return
        return $result;

    }

}